module nlcg_mod
! non-linear CG step
  use sep
  use mem_mod

  implicit none

  integer, private                                     :: n, itry, ntry
  double precision, private                            :: r0, r1, r2, r3, step_size, beta
  real, dimension(:), allocatable, private             :: model, gradient0, gradient, search, mmin, mmax

  contains

!!---------------------------------------------------------- 

  subroutine nlcg_init(n_in, ntry_in, mmin_in, mmax_in)
    integer                       :: n_in, ntry_in
    real, dimension(:)            :: mmin_in, mmax_in
    n = n_in
    ntry = ntry_in
    call mem_alloc(model, n)
    call mem_alloc(gradient0, n)
    call mem_alloc(gradient, n)
    call mem_alloc(search, n)
    call mem_alloc(mmin, n)
    call mem_alloc(mmax, n)
    mmin = mmin_in
    mmax = mmax_in
  end subroutine

!!---------------------------------------------------------- 

  subroutine nlcg_update(m_in)
    real, dimension(:)            :: m_in
    where(m_in < mmin) m_in = mmin
    where(m_in > mmax) m_in = mmax
    model = m_in
  end subroutine

!!---------------------------------------------------------- 

  subroutine nlcg(f_in, g_in, m_in, task_in)
    integer                       :: task_in, stat
    double precision              :: f_in
    real, dimension(:)            :: g_in, m_in
    if(task_in .eq. 0) then
      where(m_in < mmin) m_in = mmin
      where(m_in > mmax) m_in = mmax
      model = m_in
      gradient = 0.
      search = 0.
      itry = 0
      r0 = 0.
      r1 = 0.
      r2 = 0.
      r3 = 0.
      step_size = 0.
      beta = 0.
      task_in = 1
    else if(task_in .eq. 1) then
      r0 = f_in
      model = m_in
      gradient0 = gradient
      gradient = g_in
      if(step_size .ne. 0.) beta = sum(dprod(gradient,gradient-gradient0))/sum(dprod(gradient0,gradient0))
      beta = max(beta,0.)
      search = gradient + beta * search
      if(step_size .eq. 0.) then
        step_size = sqrt(sum(dprod(m_in,m_in))) / sqrt(sum(dprod(search,search))) / 100.
        write(0,*) "Step size is resetted to ", step_size
      end if
      itry = 0
      m_in = model - search * step_size * 0.5
      where(m_in < mmin) m_in = mmin
      where(m_in > mmax) m_in = mmax
      task_in = 2
    else if(task_in .eq. 2) then
      r1 = f_in
      m_in = model - search * step_size * 1.0
      where(m_in < mmin) m_in = mmin
      where(m_in > mmax) m_in = mmax
      task_in = 3
    else if(task_in .eq. 3) then
      r2 = f_in
      m_in = model - search * step_size * 1.5
      where(m_in < mmin) m_in = mmin
      where(m_in > mmax) m_in = mmax
      task_in = 4
    else if(task_in .eq. 4) then
      r3 = f_in
      write(0,*) "residuals =", real(r0), real(r1), real(r2), real(r3)
      call stepper(r0, r1, r2, r3, step_size, stat)
      if(stat .eq. 0) then
        model = model - search * step_size
        where(model < mmin) model = mmin
        where(model > mmax) model = mmax
        m_in = model
        task_in = 5
      else
        itry = itry + 1
        task_in = 2
        m_in = model - search * step_size * 0.5
        where(m_in < mmin) m_in = mmin
        where(m_in > mmax) m_in = mmax
        if(itry .eq. ntry) then
          step_size = sqrt(sum(dprod(m_in,m_in))) / sqrt(sum(dprod(search,search))) / 100.
          write(0,*) "Step size is resetted to ", step_size
          m_in = model - search * step_size * 0.5
          where(m_in < mmin) m_in = mmin
          where(m_in > mmax) m_in = mmax
        else if(itry .eq. 2*ntry) then
          write(0,*) "trying iterations reached limit without finding a proper step size, exit the program"
          task_in = 6
          m_in = model
        end if
      end if
    else if(task_in .eq. 5) then
      task_in = 1
    end if

  end subroutine

!!---------------------------------------------------------- 

  subroutine stepper(r0, r1, r2, r3, step_size, stat)
    !! testing step sizes
    double precision              :: r0, r1, r2, r3, step_size
    integer                       :: stat

    if(isnan(r1) .or. r1 .eq. r1 + r1) then
      write(0,*) "something wrong with r1, will reset its value"
      r1 = r0 + abs(r0)
    end if
    if(isnan(r2) .or. r2 .eq. r2 + r2) then
      write(0,*) "something wrong with r2, will reset its value"
      r2 = r0 + abs(r0)
    end if
    if(isnan(r3) .or. r3 .eq. r3 + r3) then
      write(0,*) "something wrong with r3, will reset its value"
      r3 = r0 + abs(r0)
    end if

    if(r0 < r1 .and. r0 < r2 .and. r0 < r3) then
      write(0,*) "all residuals are larger than r0, will try dividing by 4"
      step_size = step_size * 0.25
      stat = 1
    else if(r1 < r2 .and. r1 < r3) then
      step_size = step_size * 0.5
      write(0,*) "best scalar is 0.5, best step size is ", step_size
      stat = 0
    else if(r2 < r1 .and. r2 < r3) then
      write(0,*) "best scalar is 1.0, best step size is ", step_size
      stat = 0
    else if(r3 < r1 .and. r3 < r2) then
      step_size = step_size * 1.5
      write(0,*) "best scalar is 1.5, best step size is ", step_size
      stat = 0
    else
      write(0,*) "unknown step size problem, will try dividing by 4"
      step_size = step_size * 0.25
      stat = 1
    end if

  end subroutine

end module
